"""Test for Lock"""

from the_keyspy.devices.lock import TheKeysLock
from the_keyspy.devices.gateway import TheKeysGateway

import unittest


class TestLock(unittest.TestCase):
    def setUp(self) -> None:
        self.lock = TheKeysLock(1, TheKeysGateway(
            1, "0.0.0.0"), "home", "", "1234")

    def test_battery_level(self):
        self.lock._battery = 3600
        self.assertEqual(self.lock.battery_level, 0)
        self.lock._battery = 8000
        self.assertEqual(self.lock.battery_level, 100)
        self.lock._battery = 5800
        self.assertEqual(self.lock.battery_level, 50)



from dataclasses import dataclass, field
from typing import Any, Optional

from dataclasses_json import LetterCase, config, dataclass_json


# Common
@dataclass_json
@dataclass
class DatabaseActionDate:
    """Data class for timestamp"""

    date: str
    timezone_type: int
    timezone: str


@dataclass_json
@dataclass
class Info:
    last_seen: str
    ip: Optional[str]


# Accessoire
@dataclass_json
@dataclass
class Accessoire:
    """Data class for accessory"""

    id: int
    id_accessoire: str
    nom: str
    description: Optional[str]
    type: int
    version: int
    type_version: int
    created_at: DatabaseActionDate
    updated_at: DatabaseActionDate
    public_key: str
    info: Info
    configuration: list[Any]
    cfg: Any


# Serrure
@dataclass_json
@dataclass
class SerrureAccessoireAccessoire:
    """Data class for lock accessory"""

    id: int
    id_accessoire: str
    nom: str
    type: int
    version: int
    type_version: int
    info: Info
    configuration: list[Any]


@dataclass_json
@dataclass
class SerrureAccessoire:
    """Data class for lock accessory"""

    id: int
    accessoire: SerrureAccessoireAccessoire
    info: Optional[str]


@dataclass_json
@dataclass
class SerrureProduit:
    """Data class for lock product"""

    id: int
    nom: str
    version: int
    version_beta: int = field(metadata=config(letter_case=LetterCase.CAMEL))


@dataclass_json
@dataclass
class SerrureCompte:
    id: int
    nom: str


@dataclass_json
@dataclass
class SerrureLogData:
    ts: int


@dataclass_json
@dataclass
class SerrureLog:
    id: int
    action_at: DatabaseActionDate
    created_at: DatabaseActionDate
    action: str
    data: SerrureLogData
    status: int
    share_id: Optional[int]
    accessory_share_id: Optional[int]
    utilisateur: str
    utilisateur_id: Optional[int]
    accessoire_id: Optional[int]


@dataclass_json
@dataclass
class Serrure:
    """Data class for lock"""

    id: int
    id_serrure: str
    code: str
    code_serrure: str
    etat: str
    nom: str
    qrcode: str
    serrure_droite: bool
    main_libre: bool
    longitude: int
    latitude: int
    radius: int
    timezone: str
    max_speed: int = field(metadata=config(letter_case=LetterCase.CAMEL))
    latch_delay: int = field(metadata=config(letter_case=LetterCase.CAMEL))
    assisted_actions: bool = field(metadata=config(letter_case=LetterCase.CAMEL))
    unlock_only: bool = field(metadata=config(letter_case=LetterCase.CAMEL))
    description: Optional[str]
    version: int
    version_cible: int
    beta: int
    produit: SerrureProduit
    compte: SerrureCompte
    created_at: DatabaseActionDate
    updated_at: DatabaseActionDate
    log_sequence: int = field(metadata=config(letter_case=LetterCase.CAMEL))
    logs: list[SerrureLog]
    public_key: str
    message: str
    accessoires: list[SerrureAccessoire]
    battery: int
    battery_date: DatabaseActionDate
    door: int


# Utilisateur
@dataclass_json
@dataclass
class UtilisateurSerrureUtilisateur:
    username: str
    firstname: str
    lastname: str


@dataclass_json
@dataclass
class UtilisateurSerrureAccessoireAccessoire:
    id: int
    id_accessoire: str
    nom: str
    type: int
    configuration: list[Any]


@dataclass_json
@dataclass
class UtilisateurSerrureAccessoire:
    id: int
    accessoire: UtilisateurSerrureAccessoireAccessoire
    info: Optional[Info]


@dataclass_json
@dataclass
class UtilisateurSerrureProduit:
    id: int
    nom: str
    version: int
    version_beta: int = field(metadata=config(letter_case=LetterCase.CAMEL))


@dataclass_json
@dataclass
class UtilisateurSerrure:
    """Data class for utilisateur"""

    id: int
    id_serrure: str
    code: str
    code_serrure: str
    etat: str
    nom: str
    couleur: Any
    qrcode: str
    serrure_droite: bool
    main_libre: bool
    longitude: int
    latitude: int
    radius: int
    timezone: str
    max_speed: int = field(metadata=config(letter_case=LetterCase.CAMEL))
    latch_delay: int = field(metadata=config(letter_case=LetterCase.CAMEL))
    assisted_actions: bool = field(metadata=config(letter_case=LetterCase.CAMEL))
    unlock_only: bool = field(metadata=config(letter_case=LetterCase.CAMEL))
    description: Optional[str]
    log_sequence: int = field(metadata=config(letter_case=LetterCase.CAMEL))
    public_key: str
    message: str
    utilisateur: UtilisateurSerrureUtilisateur
    version: int
    version_cible: int
    beta: int
    battery: int
    battery_date: DatabaseActionDate
    accessoires: list[UtilisateurSerrureAccessoire]
    produit: UtilisateurSerrureProduit


@dataclass_json
@dataclass
class Utilisateur:
    """Data class for utilisateur"""

    id: str
    type: str
    roles: list[str]
    firstname: str
    lastname: str
    locale: str
    username: str
    email: str
    created_at: DatabaseActionDate
    updated_at: DatabaseActionDate
    notification_token: str
    notification_enabled: bool
    serrures: list[UtilisateurSerrure]
    tel: str


# Partage
@dataclass_json
@dataclass
class PartageRole:
    """Data class for share role"""

    id: int
    description: str


@dataclass_json
@dataclass
class PartageUtilisateurUtilisateur:
    """Data class for share user"""

    username: str
    prenom: str
    nom: str
    email: str
    telephone: Optional[str] = None

    def __post_init__(self):
        self.telephone = self.username if self.telephone is None else self.telephone


@dataclass_json
@dataclass
class PartageAccessoireAccessoire:
    """Data class for share accessory"""

    id: int
    id_accessoire: str
    nom: str
    type: int
    version: int
    type_version: int
    configuration: list[Any]


@dataclass_json
@dataclass
class PartageCommun:
    """Data class for abstract share"""

    id: int
    nom: str
    date_debut: Optional[str]
    date_fin: Optional[str]
    heure_debut: Optional[str]
    heure_fin: Optional[str]
    description: Optional[str]
    notification_enabled: bool
    horaires: list[Any]


@dataclass_json
@dataclass
class PartageUtilisateur(PartageCommun):
    """Data class for user share"""

    actif: bool
    role: PartageRole
    utilisateur: PartageUtilisateurUtilisateur


@dataclass_json
@dataclass
class PartageAccessoire(PartageCommun):
    """Data class for accessory share"""

    iddesc: Optional[str]
    nom: str
    actif: bool
    accessoire: PartageAccessoireAccessoire
    code: str


@dataclass_json
@dataclass
class PartageDemandeUtilisateur:
    telephone: str


@dataclass_json
@dataclass
class PartageDemande(PartageCommun):
    role: PartageRole
    code: str
    utilisateur: PartageDemandeUtilisateur


@dataclass_json
@dataclass
class Partage:
    """Data class for share"""

    partages_utilisateur: list[PartageUtilisateur]
    partages_accessoire: list[PartageAccessoire]
    partages_demande: list[PartageDemande]



import json
from typing import Union
from flask.json.provider import JSONProvider


def create_database_action_date():
    return {"date": "2020-01-01 00:00:00.000000", "timezone_type": 3, "timezone": "Europe/Berlin"}


class JSON_Improved(json.JSONEncoder):
    def default(self, o):
        return o.__dict__()


class CustomJSONProvider(JSONProvider):
    def dumps(self, obj, **kwargs):
        return json.dumps(obj, **kwargs, cls=JSON_Improved)

    def loads(self, s: Union[str, bytes], **kwargs):
        return json.loads(s, **kwargs)


class BaseMock:
    def __iter__(self):
        yield from self.__dict__.items()

    def __str__(self):
        return json.dumps(dict(self), ensure_ascii=False)

    def __repr__(self):
        return self.__str__()


class HasId:
    def __init__(self, id: int = 1, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.with_id(id)

    def with_id(self, id: int):
        self._id = id
        return self


class HasUsername:
    def __init__(self, username: str = "username", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.with_username(username)

    def with_username(self, username: str):
        self._username = username
        return self


class HasFirstname:
    def __init__(self, firstname: str = "firstname", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.with_firstname(firstname)

    def with_firstname(self, firstname: str):
        self._firstname = firstname
        return self


class HasLastname:
    def __init__(self, lastname: str = "lastname", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.with_lastname(lastname)

    def with_lastname(self, lastname: str):
        self._lastname = lastname
        return self


class UtilisateurSerrureAccessoireMock(BaseMock, HasId):
    def __init__(self, id: int = 1):
        super().__init__(id)

    def __dict__(self):
        return {
            "id": self._id,
            "accessoire": {"id": 1, "id_accessoire": "id_accessoire", "nom": "TK Gateway", "type": 1, "configuration": []},
            "info": {"last_seen": "2020-01-01 00:00:00", "ip": "192.168.1.100"},
        }


class UtilisateurSerrureMock(BaseMock, HasId, HasUsername, HasFirstname, HasLastname):
    def __init__(self, id: int = 1, username: str = "+33123456789", firstname: str = "John", lastname: str = "Doe"):
        super().__init__(id=id, username=username, firstname=firstname, lastname=lastname)
        self._accessoires: list[UtilisateurSerrureAccessoireMock] = []

    def with_accessoire(self, utilisateur_serrure_accessoire: UtilisateurSerrureAccessoireMock):
        self._accessoires.append(utilisateur_serrure_accessoire)
        return self

    def __dict__(self):
        return {
            "id": self._id,
            "id_serrure": "1",
            "code": "code",
            "code_serrure": "code_serrure",
            "etat": "open",
            "nom": "Home",
            "couleur": None,
            "qrcode": "qrcode",
            "serrure_droite": True,
            "main_libre": True,
            "longitude": 0,
            "latitude": 0,
            "radius": 100,
            "timezone": "Europe/Paris",
            "maxSpeed": 80,
            "latchDelay": 1000,
            "assistedActions": False,
            "unlockOnly": False,
            "description": None,
            "logSequence": 0,
            "public_key": "public_key",
            "message": "",
            "utilisateur": {"username": self._username, "firstname": self._firstname, "lastname": self._lastname},
            "version": 81,
            "version_cible": 81,
            "beta": 0,
            "battery": 7235,
            "battery_date": create_database_action_date(),
            "accessoires": self._accessoires,
            "produit": {"id": 1, "nom": "TK 1.5.1", "version": 81, "versionBeta": 82},
        }


class UtilisateurMock(BaseMock, HasId, HasUsername, HasFirstname, HasLastname):
    def __init__(self, id: int = 1, username: str = "+33123456789", firstname: str = "John", lastname: str = "Doe"):
        super().__init__(id=id, username=username, firstname=firstname, lastname=lastname)
        self._serrures: list[UtilisateurSerrureMock] = []

    def with_serrure(self, utilisateur_serrure: UtilisateurSerrureMock):
        utilisateur_serrure.with_username(self._username)
        self._serrures.append(utilisateur_serrure)
        return self

    def __dict__(self):
        return {
            "id": self._id,
            "type": "user_utilisateur",
            "roles": ["ROLE_UTILISATEUR"],
            "firstname": self._firstname,
            "lastname": self._lastname,
            "locale": "fr",
            "username": self._username,
            "email": "john.doe@mail.com",
            "created_at": create_database_action_date(),
            "updated_at": create_database_action_date(),
            "notification_token": "notification_token",
            "notification_enabled": True,
            "serrures": self._serrures,
            "tel": self._username,
        }


class AccessoireMock(BaseMock, HasId):
    def __dict__(self):
        return {
            "id": int(self._id),
            "id_accessoire": str(self._id),
            "nom": "TK Gateway",
            "description": None,
            "type": 1,
            "version": 65,
            "type_version": 65,
            "created_at": create_database_action_date(),
            "updated_at": create_database_action_date(),
            "public_key": "public_key",
            "info": {"last_seen": "2020-01-01 00:00", "ip": "127.0.0.1:5000"},
            "configuration": [],
            "cfg": None,
        }


class PartageUtilisateurMock(BaseMock, HasId, HasUsername, HasFirstname, HasLastname):
    def __init__(self, id: int = 1, username: str = "+33123456789", firstname: str = "John", lastname: str = "Doe"):
        super().__init__(id=id, username=username, firstname=firstname, lastname=lastname)

    def __dict__(self):
        return {
            "id": self._id,
            "nom": "John Doe",
            "actif": True,
            "role": {"id": 1, "description": "owner"},
            "date_debut": None,
            "date_fin": None,
            "heure_debut": None,
            "heure_fin": None,
            "horaires": [],
            "description": None,
            "notification_enabled": True,
            "utilisateur": {
                "username": self._username,
                "prenom": self._firstname,
                "nom": "Doe",
                "email": "john.doe@mail.com",
                "telephone": self._username,
            },
            "remoteKeySharingId": 1,
        }


class PartageAccessoireMock(BaseMock, HasId):
    def __init__(self, id: int = 1):
        super().__init__(id=id)

    def __dict__(self):
        return {
            "id": self._id,
            "iddesc": "remote",
            "nom": "TheKeysPy (Remote)",
            "actif": True,
            "date_debut": None,
            "date_fin": None,
            "heure_debut": None,
            "heure_fin": None,
            "description": None,
            "notification_enabled": True,
            "accessoire": {
                "id": 1,
                "id_accessoire": "id_accessoire",
                "nom": "TK Gateway",
                "type": 1,
                "version": 65,
                "type_version": 65,
                "configuration": [],
            },
            "horaires": [],
            "code": "code",
        }


class PartageMock(BaseMock):
    def __init__(self):
        self._partages_utilisateur: list[PartageUtilisateurMock] = []
        self._partages_accessoire: list[PartageAccessoireMock] = []

    def with_partage_utilisateur(self, partage_utilisateur: PartageUtilisateurMock):
        self._partages_utilisateur.append(partage_utilisateur)
        return self

    def with_partage_accessoire(self, partage_accessoire: PartageAccessoireMock):
        self._partages_accessoire.append(partage_accessoire)
        return self

    def __dict__(self):
        return {
            "partages_utilisateur": self._partages_utilisateur,
            "partages_accessoire": self._partages_accessoire,
            "partages_demande": [],
        }



"""The Keys lock device implementation"""
from typing import Any

from .base import TheKeysDevice
from .gateway import TheKeysGateway

OPENED = "Door open"
CLOSED = "Door closed"
JAMMED = "Door jammed"
UNKNOWN = ""


def _map(x, in_min, in_max, out_min, out_max):
    return int((x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min)


class TheKeysLock(TheKeysDevice):
    """Lock device implementation"""

    def __init__(self, id: int, gateway: TheKeysGateway, name: str, identifier: str, share_code: str) -> None:
        super().__init__(id)
        self._gateway = gateway
        self._name = name
        self._identifier = identifier
        self._share_code = share_code

        self._status = UNKNOWN
        self._code = 0
        self._version = 0
        self._position = 0
        self._rssi = 0
        self._battery = 0

    def open(self) -> bool:
        """Open this lock"""
        result = self._gateway.open(self._identifier, self._share_code)
        if result:
            self._status = OPENED
        else:
            self._status = JAMMED

        return result

    def close(self) -> bool:
        """Close this lock"""
        result = self._gateway.close(self._identifier, self._share_code)
        if result:
            self._status = CLOSED
        else:
            self._status = JAMMED

        return result

    def calibrate(self) -> bool:
        """Calibrate this lock"""
        return self._gateway.calibrate(self._identifier, self._share_code)

    def status(self) -> Any:
        """Return this lock status"""
        return self._gateway.locker_status(self._identifier, self._share_code)

    def synchronize(self) -> Any:
        return self._gateway.synchronize_locker(self._identifier)

    def update(self) -> Any:
        return self._gateway.update_locker(self._identifier)

    def retrieve_infos(self) -> None:
        json = self.status()
        if json["status"] == "ko":
            return

        self._status = json["status"]
        self._code = json["code"]
        self._version = json["version"]
        self._position = json["position"]
        self._rssi = json["rssi"]
        self._battery = json["battery"]

    @property
    def name(self) -> str:
        """This lock name"""
        return self._name

    @property
    def is_unlocked(self) -> bool:
        """Is this lock unlocked"""
        return self._status == OPENED

    @property
    def is_locked(self) -> bool:
        """Is this lock locked"""
        return self._status == CLOSED

    @property
    def is_jammed(self) -> bool:
        """Is this lock jammed"""
        return self._status == JAMMED

    @property
    def battery_level(self) -> int:
        """The battery percentage"""
        return _map(self._battery, 3600, 8000, 0, 100)



"""Test for TheKeyApi"""
import unittest
from typing import Any

from http_server_mock import HttpServerMock

from the_keyspy import TheKeysApi
from the_keyspy.errors import (
    NoAccessoriesFoundError,
    NoGatewayAccessoryFoundError,
    GatewayAccessoryNotFoundError,
    NoUtilisateurFoundError,
    NoSharesFoundError,
)
from . import (
    CustomJSONProvider,
    UtilisateurMock,
    UtilisateurSerrureMock,
    UtilisateurSerrureAccessoireMock,
    AccessoireMock,
    PartageMock,
    PartageUtilisateurMock,
    PartageAccessoireMock,
)
from flask import jsonify


def login_check():
    return jsonify(
        {
            "access_token": "access_token",
            "expires_in": 3600,
            "token_type": "bearer",
            "scope": "actions",
            "refresh_token": "refresh_token",
            "token": "token",
        }
    )


def build_response(data: Any, status: int = 200):
    return jsonify(
        {
            "status": status,
            "data": data,
            "message": {"global": [], "form": []},
        }
    )


def utilisateur_without_serrure(username: str):
    return build_response(UtilisateurMock(username=username))


def utilisateur_with_serrure_without_accessoire(username: str):
    return build_response(UtilisateurMock(username=username).with_serrure(UtilisateurSerrureMock()))


def utilisateur_with_serrure_and_gateway(username: str):
    return build_response(
        UtilisateurMock(username=username).with_serrure(
            UtilisateurSerrureMock().with_accessoire(UtilisateurSerrureAccessoireMock()))
    )


def utilisateur_with_serrure_but_no_gateway_accessory(username: str):
    serrure = UtilisateurSerrureMock()

    class NonGatewayAccessoireMock(UtilisateurSerrureAccessoireMock):
        def __dict__(self):
            return {
                "id": self._id,
                "accessoire": {"id": 1, "id_accessoire": "id_accessoire", "nom": "Other Device", "type": 2, "configuration": []},
                "info": None,
            }
    serrure.with_accessoire(NonGatewayAccessoireMock())
    return build_response(UtilisateurMock(username=username).with_serrure(serrure))


def accessoire(id: int):
    return build_response(AccessoireMock(id))


def accessoire_not_found(id: str):
    return build_response(None)


def create_partage(id_serrure: int, id_accessoire: str):
    return build_response({"id": 2, "code": "code"})


def partage_without_partages(id_serrure: int):
    return build_response(PartageMock())


def partage_with_one_partage_utilisateur(id_serrure: int):
    return build_response(PartageMock().with_partage_utilisateur(PartageUtilisateurMock()).with_partage_accessoire(PartageAccessoireMock()))


def utilisateur_not_found(username: str):
    return build_response(None)


def partage_not_found(id_serrure: int):
    return build_response(None)


def locker_status():
    return jsonify({"status": "Door open", "code": 1, "id": 1, "version": 81, "position": 20, "rssi": 0, "battery": 7235})


class TheKeyApiTest(unittest.TestCase):
    """The KeysApi test class"""

    def setUp(self):
        super().setUp()
        self.app = HttpServerMock(__name__)
        self.app.add_url_rule("/api/login_check", None,
                              view_func=login_check, methods=["POST"])
        self.app.json = CustomJSONProvider(self.app)

    def test_utilisateur_without_serrure(self):
        self.app.add_url_rule("/fr/api/v2/utilisateur/get/<username>",
                              None, view_func=utilisateur_without_serrure, methods=["GET"])
        with self.app.run("localhost", 5000):
            controller = TheKeysApi(
                "+33123456789", "password", "http://localhost:5000")
            # Un utilisateur sans serrure devrait retourner une liste vide, pas lever d'exception
            self.assertEqual(controller.get_devices(), [])

    def test_utilisateur_with_serrure_without_accessoire(self):
        self.app.add_url_rule("/fr/api/v2/utilisateur/get/<username>", None,
                              view_func=utilisateur_with_serrure_without_accessoire, methods=["GET"])
        with self.app.run("localhost", 5000):
            controller = TheKeysApi(
                "+33123456789", "password", "http://localhost:5000")
            with self.assertRaises(NoAccessoriesFoundError):
                controller.get_devices()

    def test_utilisateur_with_serrure_gateway_without_partage(self):
        self.app.add_url_rule("/fr/api/v2/utilisateur/get/<username>", None,
                              view_func=utilisateur_with_serrure_and_gateway, methods=["GET"])
        self.app.add_url_rule(
            "/fr/api/v2/accessoire/get/<id>", None, view_func=accessoire)
        self.app.add_url_rule("/fr/api/v2/partage/all/serrure/<id_serrure>",
                              None, view_func=partage_without_partages)
        self.app.add_url_rule("/fr/api/v2/partage/create/<id_serrure>/accessoire/<id_accessoire>",
                              None, view_func=create_partage, methods=["POST"])
        self.app.add_url_rule("/locker_status", None,
                              view_func=locker_status, methods=["POST"])
        with self.app.run("localhost", 5000):
            controller = TheKeysApi(
                "+33123456789", "password", "http://localhost:5000")
            self.assertEqual(len(controller.get_devices()), 2)

    def test_utilisateur_with_serrure_gateway_partage(self):
        self.app.add_url_rule("/fr/api/v2/utilisateur/get/<username>", None,
                              view_func=utilisateur_with_serrure_and_gateway, methods=["GET"])
        self.app.add_url_rule(
            "/fr/api/v2/accessoire/get/<id>", None, view_func=accessoire)
        self.app.add_url_rule("/fr/api/v2/partage/all/serrure/<id_serrure>",
                              None, view_func=partage_with_one_partage_utilisateur)
        self.app.add_url_rule("/locker_status", None,
                              view_func=locker_status, methods=["POST"])
        with self.app.run("localhost", 5000):
            controller = TheKeysApi(
                "+33123456789", "password", "http://localhost:5000")
            self.assertEqual(len(controller.get_devices()), 2)

    def test_utilisateur_with_serrure_but_no_gateway_accessory(self):
        self.app.add_url_rule("/fr/api/v2/utilisateur/get/<username>", None,
                              view_func=utilisateur_with_serrure_but_no_gateway_accessory, methods=["GET"])
        with self.app.run("localhost", 5000):
            controller = TheKeysApi(
                "+33123456789", "password", "http://localhost:5000")
            with self.assertRaises(NoGatewayAccessoryFoundError):
                controller.get_devices()

    def test_gateway_accessoire_not_found(self):
        def utilisateur_with_gateway_but_accessoire_not_found(username: str):
            return build_response(UtilisateurMock(username=username).with_serrure(UtilisateurSerrureMock().with_accessoire(UtilisateurSerrureAccessoireMock())))

        def accessoire_not_found(id: str):
            return build_response(None)
        self.app.add_url_rule("/fr/api/v2/utilisateur/get/<username>", None,
                              view_func=utilisateur_with_gateway_but_accessoire_not_found, methods=["GET"])
        self.app.add_url_rule("/fr/api/v2/accessoire/get/<id>",
                              None, view_func=accessoire_not_found)
        with self.app.run("localhost", 5000):
            controller = TheKeysApi(
                "+33123456789", "password", "http://localhost:5000")
            with self.assertRaises(GatewayAccessoryNotFoundError):
                controller.get_devices()

    def test_utilisateur_not_found(self):
        self.app.add_url_rule("/fr/api/v2/utilisateur/get/<username>", None,
                              view_func=utilisateur_not_found, methods=["GET"])
        with self.app.run("localhost", 5000):
            controller = TheKeysApi(
                "+33123456789", "password", "http://localhost:5000")
            with self.assertRaises(NoUtilisateurFoundError):
                controller.get_devices()

    def test_partage_not_found(self):
        self.app.add_url_rule("/fr/api/v2/utilisateur/get/<username>", None,
                              view_func=utilisateur_with_serrure_and_gateway, methods=["GET"])
        self.app.add_url_rule(
            "/fr/api/v2/accessoire/get/<id>", None, view_func=accessoire)
        self.app.add_url_rule("/fr/api/v2/partage/all/serrure/<id_serrure>", None,
                              view_func=partage_not_found)
        with self.app.run("localhost", 5000):
            controller = TheKeysApi(
                "+33123456789", "password", "http://localhost:5000")
            with self.assertRaises(NoSharesFoundError):
                controller.get_devices()



class TheKeysApiError(Exception):
    """Base exception for TheKeysApi errors."""
    pass


class NoAccessoriesFoundError(TheKeysApiError):
    """Raised when no accessories are found for a lock."""
    pass


class NoGatewayAccessoryFoundError(TheKeysApiError):
    """Raised when no gateway accessory is found for a lock."""
    pass


class GatewayAccessoryNotFoundError(TheKeysApiError):
    """Raised when the gateway accessory could not be retrieved from the API."""
    pass


class NoGatewayIpFoundError(TheKeysApiError):
    """Raised when no gateway IP is found for a lock."""
    pass


class NoSharesFoundError(TheKeysApiError):
    """Raised when no shares are found for a lock."""
    pass


class NoUtilisateurFoundError(TheKeysApiError):
    """Raised when the user could not be retrieved from the API."""
    pass



"""The Keys CLI application"""
import argparse
import logging

from src.the_keyspy import Action, TheKeysApi

parser = argparse.ArgumentParser(description="The Keys CLI")
parser.add_argument("-t", dest="telephone", help="login", required=True)
parser.add_argument("-p", dest="password", help="password", required=True)
parser.add_argument("-a", dest="action", help="action",
                    required=True, type=Action, choices=list(Action))
parser.add_argument("-d", dest="device", help="device", required=False)
parser.add_argument("--log-level", choices=["DEBUG", "INFO", "WARNING",
                    "ERROR", "CRITICAL"], default="INFO", help="log level (default: INFO)")
args = parser.parse_args()

logging.basicConfig(level=args.log_level,
                    format="%(asctime)s - %(levelname)s - %(message)s")
with TheKeysApi(args.telephone, args.password) as api:
    for device in api.get_locks():
        match args.action:
            case Action.OPEN:
                result = device.open()
            case Action.CLOSE:
                result = device.close()
            case Action.CALIBRATE:
                result = device.calibrate()
            case Action.LOCKER_STATUS:
                result = device.status()
            case Action.SYNCHRONIZE_LOCKER:
                result = device.synchronize()
            case Action.UPDATE_LOCKER:
                result = device.update()
            case Action.STATUS:
                result = device.status()
            case _:
                result = None

        print(result)



from curses import meta
from .base import TheKeysDevice
import base64
import hmac
import time
import requests
from enum import Enum
from typing import Any, Optional

import logging

logger = logging.getLogger("the_keyspy.devices.gateway")


class Action(Enum):
    """All available actions"""

    OPEN = "open"
    CLOSE = "close"
    CALIBRATE = "calibrate"
    LOCKER_STATUS = "locker_status"
    SYNCHRONIZE_LOCKER = "synchronize_locker"
    UPDATE_LOCKER = "update_locker"
    STATUS = "status"
    UPDATE = "update"

    def __str__(self):
        return self.value


class TheKeysGateway(TheKeysDevice):
    """Gateway device implementation"""

    def __init__(self, id: int, host: str) -> None:
        super().__init__(id)
        self._host = host

    def open(self, identifier: str, share_code: str) -> bool:
        return self.action(Action.OPEN, identifier, share_code)["status"] == "ok"

    def close(self, identifier: str, share_code: str) -> bool:
        return self.action(Action.CLOSE, identifier, share_code)["status"] == "ok"

    def calibrate(self, identifier: str, share_code: str) -> bool:
        return self.action(Action.CALIBRATE, identifier, share_code)["status"] == "ok"

    def locker_status(self, identifier: str, share_code: str) -> Any:
        return self.action(Action.LOCKER_STATUS, identifier, share_code)

    def synchronize_locker(self, identifier: str) -> bool:
        return self.action(Action.SYNCHRONIZE_LOCKER, identifier)["status"] == "ok"

    def update_locker(self, identifier: str) -> bool:
        return self.action(Action.UPDATE_LOCKER, identifier)["status"] == "ok"

    def status(self) -> Any:
        return self.action(Action.STATUS)

    def update(self) -> Any:
        return self.action(Action.UPDATE)

    def action(self, action: Action, identifier: str = "", share_code: str = "") -> Any:
        data = {}
        if identifier != "":
            data["identifier"] = identifier

        if share_code != "":
            timestamp = str(int(time.time()))
            data["ts"] = timestamp
            data["hash"] = base64.b64encode(hmac.new(share_code.encode(
                "ascii"), timestamp.encode("ascii"), "sha256").digest())

        match action:
            case Action.OPEN:
                url = "open"
            case Action.CLOSE:
                url = "close"
            case Action.CALIBRATE:
                url = "calibrate"
            case Action.LOCKER_STATUS:
                url = "locker_status"
            case Action.SYNCHRONIZE_LOCKER:
                url = "locker/synchronize"
            case Action.UPDATE_LOCKER:
                url = "locker/update"
            case Action.UPDATE:
                url = "update"
            case Action.STATUS:
                url = "status"
            case _:
                url = "status"

        response_data = self.__http_request(url, data)
        if "status" not in response_data:
            response_data["status"] = "ok"

        if response_data["status"] == "ko":
            raise RuntimeError(response_data)

        return response_data

    def __http_request(self, url: str, data: Optional[dict] = None) -> Any:
        method = "post" if data else "get"
        logger.debug("%s %s", method.upper(), url)
        try:
            with requests.Session() as session:
                full_url = f"http://{self._host}/{url}"
                response = session.post(
                    full_url, data=data) if method == "post" else session.get(full_url)
                logger.debug("response_data: %s", response.json())
                return response.json()
        except ConnectionError as error:
            raise error



"""
Python library to handle the keys api
"""
import logging
from typing import Any, List, TypeVar, Type

import requests

from .dataclasses import Accessoire, Partage, PartageAccessoire, Utilisateur, UtilisateurSerrureAccessoireAccessoire
from .devices import TheKeysDevice, TheKeysGateway, TheKeysLock
from .errors import (
    NoGatewayIpFoundError,
    NoUtilisateurFoundError,
    NoAccessoriesFoundError,
    NoGatewayAccessoryFoundError,
    GatewayAccessoryNotFoundError,
    NoSharesFoundError,
)

logger = logging.getLogger("the_keyspy")

BASE_URL = "https://api.the-keys.fr"
SHARE_NAME = "TheKeysPy (Remote)"
ACCESSORY_GATEWAY = 1

T = TypeVar('T')


def deserialize_dataclass(cls: Type[T], data: Any) -> T:
    """Helper function to deserialize dataclass from dict"""
    if data is None:
        raise ValueError("Cannot deserialize None data")
    return cls.from_dict(data)  # type: ignore


class TheKeysApi:
    """TheKeysApi class"""

    def __init__(self, username: str, password: str, base_url=BASE_URL) -> None:
        self._username = username
        self._password = password
        self._base_url = base_url
        self._access_token = None

    @property
    def authenticated(self):
        """Get the token"""
        return not self._access_token is None

    def find_utilisateur_by_username(self, username: str) -> Utilisateur:
        """Return user matching the passed username"""
        response_data = self.__http_get(f"utilisateur/get/{username}")["data"]
        if response_data is None:
            raise NoUtilisateurFoundError(
                "User could not be retrieved from the API.")
        return deserialize_dataclass(Utilisateur, response_data)

    def find_accessoire_by_id(self, id: int) -> Accessoire:
        """Return accessory matching the passed id"""
        response_data = self.__http_get(f"accessoire/get/{id}")["data"]
        if response_data is None:
            raise GatewayAccessoryNotFoundError(
                "Gateway accessory could not be retrieved from the API.")
        return deserialize_dataclass(Accessoire, response_data)

    def find_partage_by_lock_id(self, lock_id: int) -> Partage:
        """Return share matching the passed lock_id"""
        response_data = self.__http_get(
            f"partage/all/serrure/{lock_id}")["data"]
        if response_data is None:
            raise NoSharesFoundError(
                "No shares found for this lock.")
        return deserialize_dataclass(Partage, response_data)

    def create_accessoire_partage_for_serrure_id(
        self, serrure_id: int, share_name: str, accessoire: UtilisateurSerrureAccessoireAccessoire
    ) -> PartageAccessoire:
        """Create a share for the passed serrure_id and accessoire"""
        data = {}
        data["partage_accessoire[description]"] = ""
        data["partage_accessoire[nom]"] = share_name
        data["partage_accessoire[iddesc]"] = "remote"

        response = self.__http_post(
            f"partage/create/{serrure_id}/accessoire/{accessoire.id_accessoire}", data)["data"]
        partage_accessoire = {}
        partage_accessoire["id"] = response["id"]
        partage_accessoire["iddesc"] = "remote"
        partage_accessoire["nom"] = share_name
        partage_accessoire["actif"] = True
        partage_accessoire["date_debut"] = None
        partage_accessoire["date_fin"] = None
        partage_accessoire["heure_debut"] = None
        partage_accessoire["heure_fin"] = None
        partage_accessoire["description"] = None
        partage_accessoire["notification_enabled"] = True
        partage_accessoire["accessoire"] = accessoire
        partage_accessoire["horaires"] = []
        partage_accessoire["code"] = response["code"]
        return deserialize_dataclass(PartageAccessoire, partage_accessoire)

    def get_locks(self) -> List[TheKeysLock]:
        return list(device for device in self.get_devices() if isinstance(device, TheKeysLock))

    def get_gateways(self) -> List[TheKeysGateway]:
        return list(device for device in self.get_devices() if isinstance(device, TheKeysGateway))

    def get_devices(self, share_name=SHARE_NAME) -> List[TheKeysDevice]:
        """Return all devices"""
        devices = []
        user = self.find_utilisateur_by_username(self._username)
        for serrure in user.serrures:
            if not serrure.accessoires:
                raise NoAccessoriesFoundError(
                    "No accessories found for this lock.")

            accessoire = next((x for x in serrure.accessoires if x.accessoire.type ==
                              ACCESSORY_GATEWAY and x.info is not None and x.info.ip is not None), None)
            if not accessoire:
                raise NoGatewayAccessoryFoundError(
                    "No gateway accessory found for this lock.")

            gateway_accessoire = self.find_accessoire_by_id(
                accessoire.accessoire.id)
            if not gateway_accessoire:
                raise GatewayAccessoryNotFoundError(
                    "Gateway accessory could not be retrieved from the API.")

            ip = gateway_accessoire.info.ip or (
                accessoire.info.ip if accessoire.info else None)
            if ip is None:
                raise NoGatewayIpFoundError(
                    "No gateway IP found for this lock.")

            gateway = TheKeysGateway(gateway_accessoire.id, ip)
            devices.append(gateway)

            partages_accessoire = self.find_partage_by_lock_id(
                serrure.id).partages_accessoire
            if not partages_accessoire:
                partages_accessoire = []

            partage = next((x for x in partages_accessoire if x.nom ==
                           SHARE_NAME and x.accessoire.id == accessoire.accessoire.id), None)
            if partage is None:
                partage = self.create_accessoire_partage_for_serrure_id(
                    serrure.id, share_name, accessoire.accessoire)

            devices.append(TheKeysLock(serrure.id, gateway,
                           serrure.nom, serrure.id_serrure, partage.code))

        return devices

    def __http_request(self, method: str, url: str, data: Any = None):
        if not self.authenticated:
            self.__authenticate()

        full_url = f"{self._base_url}/fr/api/v2/{url}"
        headers = {"Authorization": f"Bearer {self._access_token}"}

        logger.debug("%s %s", method.upper(), full_url)
        if method.lower() == "get":
            response = requests.get(full_url, headers=headers)
        elif method.lower() == "post":
            response = requests.post(full_url, headers=headers, data=data)
        else:
            raise ValueError(f"HTTP method non supportée : {method}")

        if response.status_code != 200:
            raise RuntimeError(response.text)

        json_data = response.json()
        logger.debug("response_data: %s", json_data)
        return json_data

    def __http_get(self, url: str):
        return self.__http_request("get", url)

    def __http_post(self, url: str, data: Any):
        return self.__http_request("post", url, data)

    def __authenticate(self):
        response = requests.post(
            f"{self._base_url}/api/login_check",
            data={"_username": self._username, "_password": self._password},
        )

        if response.status_code != 200:
            raise RuntimeError(response.text)

        json = response.json()
        self._access_token = json["access_token"]
        self.expires_in = json["expires_in"]

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        if exception_type is not None:
            print(exception_type, exception_value)

        return True
